{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## TP n°6: analyse expressions arithmétiques en notation polonaise inverse\n",
    "\n",
    "### 1. Implémentation d'une pile \n",
    "\n",
    "Proposer une implémentation au moyen d'une liste des quatre primitives indispensables pour manipuler une pile: cree_pile(), vide(), empiler(), depiler(). La définition fonctionnelle des primitives peut être déduite des exemples suivants: \n",
    "\n",
    "```\n",
    "print(depiler(empiler(empiler(empiler(cree_pile(),3),5),7)))\n",
    "7\n",
    "print(vide(empiler(cree_pile(),5)))\n",
    "False\n",
    "```\n",
    "\n",
    "### 2. Evaluation d'une formule en notation polonaise inverse\n",
    "\n",
    "Ecrire un programme qui prend en entrée une expression arithmétique en <a href = \"https://fr.wikipedia.org/wiki/Notation_polonaise_inverse\">notation polonaise inverse</a> (sous la forme d'une chaîne de caractère), et qui donne en sortie la valeur de l'expression. L'algorithme utilisé repose sur l'idée que chaque opérande rencontré est empilé, et que chaque fois qu'on rencontre un opérateur, on dépile deux arguments (les deux derniers, donc), on calcule le résultat, et on l'empile. Si l'expression est bien formée, la valeur de l'expression se trouve au sommet de la pile à la fin du calcul.\n",
    "\n",
    "Exemples : \n",
    "```\n",
    "3 12 3 - + = 12\n",
    "7 7 * = 49\n",
    "23 12 - 8 + 2 * = 38\n",
    "```\n",
    "\n",
    "Il n'est pas demandé de vérifier que l'expression est bien formée (mais on peut essayer d'ajouter une vérification pour éviter les erreurs). Noter que si l'on veut ajouter des symboles unaires (le signe \"moins\" dans l'expression \"-3\") on doit utiliser des symboles différents de leur correspondant binaire.\n",
    "\n",
    "\n",
    "### 3. Traduction d'une formule arithmétique\n",
    "\n",
    "Ecrire un programme qui prend en entrée une expression arithmétique en notation polonaise inverse, et donne en sortie une chaîne de caractère correspondant à l'expression équivalente en notation infixe. La pile définie plus haut devrait pouvoir être utilisée.\n",
    "\n",
    "Exemples: \n",
    "```\n",
    "3 12 3 - + = (3 + (12 - 3))\n",
    "7 7 * = (7 * 7)\n",
    "23 12 - 8 + 2 * = (((23 - 12) + 8) * 2)\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 1. Implémentation d'une pile\n",
    "On propose une version simple (non objet) des quatre primitives. Il n'y a pas de contrôle d'erreur (par exemple on pourrait vérifier que la pile est bien définie, que c'est une liste, que la valeur est bien du bon type, et renvoyer un message d'erreur si on tente de dépiler une pile vide). \n",
    "\n",
    "On propose ensuite une version \"objet\", qui permet une manipulation un peu plus aisée, mais qui devrait aussi comprendre des contrôles pour éviter les cas d'erreur."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "def cree_pile():\n",
    "    return []\n",
    "\n",
    "def vide(p):\n",
    "    return p == [] # or p == None\n",
    "\n",
    "def empiler(p,val):\n",
    "    p.append(val)\n",
    "    return p\n",
    "\n",
    "def depiler(p):\n",
    "    if not vide(p):\n",
    "        return p.pop()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "7\n",
      "False\n"
     ]
    }
   ],
   "source": [
    "# Quelques tests des primitives\n",
    "print(depiler(empiler(empiler(empiler(cree_pile(),3),5),7)))\n",
    "print(vide(empiler(cree_pile(),5)))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "# version \"objet\" de la pile (minimale)\n",
    "# Aucun contrôle d'erreur dans les méthodes\n",
    "class maPile:\n",
    "    def __init__(self):\n",
    "        self._l = []\n",
    "\n",
    "    def empile(self, val):\n",
    "        self._l.append(val)\n",
    "\n",
    "    def depile(self):\n",
    "        return self._l.pop()\n",
    "\n",
    "    def vide(self):\n",
    "        return len(self._l) == 0"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 2. Interprétation d'une formule en NPI\n",
    "On propose deux fonctions différentes qui utilisent respectivement le premier jeu de primitive et la version objet."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Interprétation d'une expression en NPI \n",
    "def interpretation(expr):\n",
    "    p = cree_pile()\n",
    "    for token in expr.split():\n",
    "        if token == '+':\n",
    "            empiler(p,depiler(p)+depiler(p))\n",
    "        elif token == '*':\n",
    "            empiler(p,depiler(p)*depiler(p))\n",
    "        elif token == '-':\n",
    "            op2 = depiler(p)\n",
    "            op1 = depiler(p)\n",
    "            empiler(p, op1 - op2)\n",
    "        else:\n",
    "            empiler(p,int(token))\n",
    "    return depiler(p)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "def interpretation_oo(expr):\n",
    "    p = maPile()\n",
    "    for token in expr.split():\n",
    "        if token == '+':\n",
    "            p.empile(p.depile()+p.depile())\n",
    "        elif token == '*':\n",
    "            p.empile(p.depile()*p.depile())\n",
    "        elif token == '-':\n",
    "            op2 = p.depile()\n",
    "            op1 = p.depile()\n",
    "            p.empile(op1 - op2)\n",
    "        else:\n",
    "            p.empile(int(token))\n",
    "    return p.depile()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3 12 3 - + = 12\n",
      "7 7 * = 49\n",
      "23 12 - 8 + 2 * = 38\n"
     ]
    }
   ],
   "source": [
    "le = ['3 12 3 - +', '7 7 *', '23 12 - 8 + 2 *']\n",
    "for e in le:\n",
    "    print(\"%s = %d\" % (e,interpretation(e)))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3 12 3 - + = 12\n",
      "7 7 * = 49\n",
      "23 12 - 8 + 2 * = 38\n"
     ]
    }
   ],
   "source": [
    "le = ['3 12 3 - +', '7 7 *', '23 12 - 8 + 2 *']\n",
    "for e in le:\n",
    "    print(\"%s = %d\" % (e,interpretation_oo(e)))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 3. Traduction d'une expression en NPI vers le format infixe parenthésé.\n",
    "On utilise cette fois la pile pour stocker des chaînes de caractères, l'algorithme est sensiblement le même que pour l'interprétation, la pile jouant encore une fois un rôle majeur.\n",
    "\n",
    "Ici on exploite le fait que nous n'avons pas défini le type des valeurs qu'on empile sur la pile: on l'avait utilisée pour des entiers, mais il est possible de l'utiliser pour des str. Avec un contrôle de type plus strict, on aurait pu être amené à définir deux classes différentes, les piles d'entiers, et les piles de str."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "def traduction(expr):\n",
    "    trad = \"\"\n",
    "    p = cree_pile()              # la pile telle qu'elle est définie peut stocker des str\n",
    "    for token in expr.split():\n",
    "        if token in ['+', '-', '*', '/']:\n",
    "            op2 = depiler(p)\n",
    "            op1 = depiler(p)\n",
    "            empiler(p, \"(%s %s %s)\" % (op1,token,op2))\n",
    "        else:\n",
    "            empiler(p,token)\n",
    "    return depiler(p)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3 12 3 - + = (3 + (12 - 3))\n",
      "7 7 * = (7 * 7)\n",
      "23 12 - 8 + 2 * = (((23 - 12) + 8) * 2)\n"
     ]
    }
   ],
   "source": [
    "le = ['3 12 3 - +', '7 7 *', '23 12 - 8 + 2 *']\n",
    "for e in le:\n",
    "    print(\"%s = %s\" % (e,traduction(e)))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
